! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: fitsstat_mod
!
! Authors:  Masaaki Sakano - SSC/LUX
!
! Description: Print statistical information of a given FITS to STDOUT.
!
! Comments:
!
! TODO:
!   + Support different coordinate units, such as, physical.
!   + Support an external mask file.
!   + Support 1-dimensional table.
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module fitsstat_mod

  use types

  use param
  use dal
  use errorHandling

  use array_utils	! for type(aryStatInfoSingleT), type(aryStatInfoFlagT), printAryStatInfo()
  use dal_aux		! for getDataType() (and inside from array_utils)
  use fits_utils	! for getFitsImgStatInfoSingle()

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    setTabName
  character(64) ::&
    typeName,&	!(image|column)
    colName,&
    coordsUnit,&	!(image|physical|fk5|fk4|gal)
    regionStyle
  logical ::&
    argWithRegion,&
    withMinCoords,&
    withMaxCoords,&
    withValLower,&
    withValUpper
  real ::&
    valLower,&
    valUpper
  real, allocatable ::&
    minAreaCoords(:),&
    maxAreaCoords(:)
  real ::&
    regCentreCoords(3),&	! Up to 3-dimensional (for future use)
    regRadiiInner(2),&			! For ellipses (for future use)
    regRadiiOuter(2)			! For ellipses (for future use)
  integer ::&
    argExtNum

  ! Command-line-related global variables:
  character(255) ::&
    setName,&
    tabName
  integer(int32), allocatable ::&
    minAreaIndices(:),&
    maxAreaIndices(:)
  ! logical ::&
  !   gWithRadiiInner

  !type(aryStatInfoFlagT) ::&
  !  flagFromComArgs

  public fitsstat

contains 

  !---------------------------------------------------------------------
  subroutine fitsstat
  !---------------------------------------------------------------------

    implicit none

    ! type(aryStatInfoSingleT) :: aryStatInfo	! defined in array_utils
    type(aryStatInfoFlagT) :: flagFromComArgs
    real(double) :: valLowerDbl, valUpperDbl
    integer :: iType
    character(255) :: strType
    logical, allocatable :: arRegMask(:,:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    flagFromComArgs%isAreaLowerUsed = withMinCoords
    flagFromComArgs%isAreaUpperUsed = withMaxCoords
    flagFromComArgs%isValLowerUsed  = withValLower
    flagFromComArgs%isValUpperUsed  = withValUpper

    valLowerDbl = real(valLower, double)
    valUpperDbl = real(valUpper, double)

    if (argWithRegion) then
      if ( (trim(regionStyle) == 'circle') .or. &
           (trim(regionStyle) == 'annulus')) then
        if (argExtNum < 0 ) then
          arRegMask = getAnnularMask(setName, regCentreCoords(1), regCentreCoords(2) &
             , regRadiiOuter(1), regRadiiInner(1))	! defined in fits_utils
        else
          arRegMask = getAnnularMask(setName, regCentreCoords(1), regCentreCoords(2) &
             , regRadiiOuter(1), regRadiiInner(1), extNum=argExtNum)	! defined in fits_utils
        end if
      end if	! if ( (trim(regionStyle) == 'circle') .or. &
    end if	! if (argWithRegion) then


    select case(trim(typeName))
    case ('image')
      iType = getDataType(trim(setName))	! ssclib/dal_aux
      ! iType = getDataType(trim(setTabName), argExtNum)   ! argExtNum for the image is not supported, yet.
      select case(iType)
      case(REAL64)
        if (argWithRegion) then
          call printAryStatInfo( getFitsImgStatInfoDouble(setName &
             , arMaskIn=arRegMask &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsImgStatInfoDouble(setName &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if


      case(REAL32)
        if (argWithRegion) then
          call printAryStatInfo( getFitsImgStatInfoSingle(setName &
             , arMaskIn=arRegMask &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLower, valUpper=valUpper &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsImgStatInfoSingle(setName &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLower, valUpper=valUpper &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case(INTEGER32)
        if (argWithRegion) then
          call printAryStatInfo( getFitsImgStatInfoInt32(setName &
             , arMaskIn=arRegMask &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsImgStatInfoInt32(setName &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case(INTEGER16)
        if (argWithRegion) then
          call printAryStatInfo( getFitsImgStatInfoInt16(setName &
             , arMaskIn=arRegMask &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsImgStatInfoInt16(setName &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case(INTEGER8)
        if (argWithRegion) then
          call printAryStatInfo( getFitsImgStatInfoInt8(setName &
             , arMaskIn=arRegMask &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsImgStatInfoInt8(setName &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case default	! Either UINT* or BOOLEAN
        call getTypeName(iType, strType)
        call warning('convertToDouble', "The datatype of the input is "&
          //trim(strType)//&
          ", neither normal Integer nor Real, possibly &
          &unsigned Integer or Boolean.  In calculating its statistical &
          &information, it is converted into Double.  That should still give &
          &the reasonably accurate answer, i.e., in the precision of Double.")

        if (argWithRegion) then
          call printAryStatInfo( getFitsImgStatInfoDouble(setName &
             , arMaskIn=arRegMask &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsImgStatInfoDouble(setName &
             , minAreaIndices=minAreaIndices, maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if
  
      end select	! select case(iType)


    case ('column')	! select case(trim(typeName))
      iType = getDataType(trim(setTabName), argExtNum, colName=colName)	! ssclib/dal_aux
      select case(iType)
      case(REAL64)
	! Note: this if-clause is actually not essential
	!   and just the second part (after else) should be enough.
	!   But I am just playing safe.
        if (argExtNum < 0 ) then
          call printAryStatInfo( getFitsVecStatInfoDouble(setTabName, colName &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsVecStatInfoDouble(setTabName, colName &
             , extNum=argExtNum &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case(REAL32)
        if (argExtNum < 0 ) then
          call printAryStatInfo( getFitsVecStatInfoSingle(setTabName, colName &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLower, valUpper=valUpper &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsVecStatInfoSingle(setTabName, colName &
             , extNum=argExtNum &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLower, valUpper=valUpper &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case(INTEGER32)
        if (argExtNum < 0 ) then
          call printAryStatInfo( getFitsVecStatInfoInt32(setTabName, colName &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsVecStatInfoInt32(setTabName, colName &
             , extNum=argExtNum &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case(INTEGER16)
        if (argExtNum < 0 ) then
          call printAryStatInfo( getFitsVecStatInfoInt16(setTabName, colName &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsVecStatInfoInt16(setTabName, colName &
             , extNum=argExtNum &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case(INTEGER8)
        if (argExtNum < 0 ) then
          call printAryStatInfo( getFitsVecStatInfoInt8(setTabName, colName &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsVecStatInfoInt8(setTabName, colName &
             , extNum=argExtNum &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if

      case default	! Either UINT* or BOOLEAN
        call getTypeName(iType, strType)
        call warning('convertToDouble', "The datatype of the input is "&
          //trim(strType)//&
          ", neither normal Integer nor Real, possibly &
          &unsigned Integer or Boolean.  In calculating its statistical &
          &information, it is converted into Double.  That should still give &
          &the reasonably accurate answer, i.e., in the precision of Double.")

        if (argExtNum < 0 ) then
          call printAryStatInfo( getFitsVecStatInfoDouble(setTabName, colName &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        else
          call printAryStatInfo( getFitsVecStatInfoDouble(setTabName, colName &
             , extNum=argExtNum &
             , minAreaIndices=minAreaIndices &
             , maxAreaIndices=maxAreaIndices &
             , valLower=valLowerDbl, valUpper=valUpperDbl &
             , flagInfo=flagFromComArgs) )	! in ssclib/array_utils and fits_utils
        end if
  
      end select	! select case( getDataType(trim(setName)) )

    case default	! select case(trim(typeName))
      call error('invalidDataType','bug')	! Should not come here by any means.
    end select		! select case(trim(typeName))

    if ((argWithRegion) .and. allocated(arRegMask)) then
       deallocate(arRegMask)
    end if

  end subroutine fitsstat


  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: rankDim
    integer :: i, dimRadiiInner, dimRadiiOuter
    logical :: noColon

    !-------------------------------------------------------------------
    ! Start

    setTabName  = stringParameter('set')
    call splitSetTabName(trim(setTabName), setName, tabName, noColonFound=noColon, useBlock=.false.)

    argExtNum  = intParameter('ext')

    typeName = stringParameter('datatype')
    select case(trim(typeName))
    case ('image')
      rankDim=2
    case ('column')
      rankDim=1
      colName  = stringParameter('column')
      if (argExtNum < 0 .and. trim(tabName) == '') then
        call error('negativeExt','Either table name (in the form of set:TableName) &
           &or positive ext must be specified.')
      end if
    case default
      call error('invalidDataType','')
    end select

    allocate(minAreaCoords(rankDim),  maxAreaCoords(rankDim) &
       ,     minAreaIndices(rankDim), maxAreaIndices(rankDim))

    coordsUnit = stringParameter('coordsunit')
    argWithRegion = booleanParameter('withregion')
    withMinCoords = booleanParameter('withmincoords')
    withMaxCoords = booleanParameter('withmaxcoords')
    withValLower = booleanParameter('withvallower')
    withValUpper = booleanParameter('withvalupper')

    !! NOTE:
    ! It seems that flagFromComArgs can not be declared as a global
    ! variable and be used here, as it would cause an error in compiling:
    !    Error: fitsstat_mod.f90, line 175: Default initialised module variable FLAGFROMCOMARGS must be SAVEd
    !flagFromComArgs%isAreaLowerUsed = withMinCoords
    !flagFromComArgs%isAreaUpperUsed = withMaxCoords
    !flagFromComArgs%isValLowerUsed  = withValLower
    !flagFromComArgs%isValUpperUsed  = withValUpper

    ! gWithRadiiInner = .false.
    if (argWithRegion) then
      regionStyle = stringParameter('regionstyle')

      ! Common for all the region styles (so far, circle or annulus only).

      regCentreCoords(:) = -9.999999e33		! Initialisation
      regRadiiOuter(:)   = -1			! Initialisation
      regRadiiInner(:)   = -1			! Initialisation
      dimRadiiOuter = parameterCount('radii')

      if ( (trim(regionStyle) == 'circle') .or. &
           (trim(regionStyle) == 'annulus')) then

        ! Basic parameter checks.
        if (rankDim .ne. parameterCount('centrecoords')) then
          call error('wrongParamNumberCentreCoords', '')
        else if ((dimRadiiOuter < 1) .or. (dimRadiiOuter > 2)) then
          call error('wrongParamNumberRadii', '')
        end if

        do i=1,rankDim	! rankDim maybe (in future) up to 3 (for cube).
          regCentreCoords(i) = realParameter('centrecoords', i-1)
          if ((i <= size(regRadiiOuter)) .and. (i <= dimRadiiOuter)) then
            regRadiiOuter(i) = realParameter('radii', i-1)
          end if
        end do

        dimRadiiInner = parameterCount('radiiinner')
        if (dimRadiiInner > 0) then
          if (dimRadiiOuter .ne. dimRadiiInner) then
            call error('wrongParamNumberRadiiInner', '')
          end if

          do i=1,rankDim	! rankDim maybe (in future) up to 3 (for cube).
            if ((i <= size(regRadiiInner)) .and. (i <= dimRadiiInner)) then
              regRadiiInner(i) = realParameter('radiiinner', i-1)
              if (regRadiiInner(i) > regRadiiOuter(i)) then
                call error('largerInnerRadiusThanOuter', '')
              end if
            end if
          end do
        end if	! if (dimRadiiInner > 0) then

      else
        call error('invalidRegionStyle', '')
      end if	! if ( (trim(regionStyle) == 'circle') .or. &
    end if	! if (argWithRegion) then

    if (withMinCoords) then
      if (rankDim .ne. parameterCount('minareacoords')) then
        call error('wrongParamNumberMinAreaCoords', '')
      end if
      do i=1,rankDim
        minAreaCoords(i) = realParameter('minareacoords', i-1)
      end do
      minAreaIndices = int(minAreaCoords, int32)	! check with 'coordsunit'
    end if

    if (withMaxCoords) then
      if (rankDim .ne. parameterCount('maxareacoords')) then
        call error('wrongParamNumberMaxAreaCoords', '')
      end if
      do i=1,rankDim
        maxAreaCoords(i) = realParameter('maxareacoords', i-1)
      end do
      maxAreaIndices = int(maxAreaCoords, int32)	! check with 'coordsunit'
    end if

    if (withMinCoords .and. withMaxCoords) then
      do i=1,rankDim
        if (minAreaCoords(i) > maxAreaCoords(i)) then
          call error('inconsistentMinMaxAreaCoords' &
             , 'An element(s) in minareacoords is larger than that in maxareacoords.')
        end if
      end do
    end if

    if (withValLower) valLower = realParameter('vallower')
    if (withValUpper) valUpper = realParameter('valupper')

  end subroutine readParameters


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fitsstat_mod
